// Module: ImageRegion
// Author: Richard Saxton
//
// Description: 
//
// This class defines a section of an image and includes methods
// to manioulate it.
//
// Modifications: See ChangeLog
// 
// $Id: ImageRegion.cc,v 1.19 2009/07/21 12:36:59 rsaxton Exp $

// General classes
#include <cmath>

// SAS-specific classes
#include "errstr.h"
#include "Tracer.h"
#include "stringutil.h"
#include "PixelIterator.h"
#include "ImagePixelIterator.h"
#include "TablePixelIterator.h"
#include "SourcePosition.h"
#include "DatafileAttributes.h"
#include "PosWCS.h"
#include "Xmm.h"
#include "DataFilter.h"
#include "DssBuilder.h"
#include "DataSubspace.h"
#include "Evaluator.h"
#include "metatask.h"
#include "attcalc.h"
#include "CalServer.h"


// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "NumericalRange.h"

// Local classes
#include "ImageRegion.h"


/* Constructor */
ImageRegion::ImageRegion( const string & imagefile, const string & expression,
      const string & tempfilename):
  _expr(expression),
  _filteredtablename(tempfilename),
  _imagefile(imagefile),
  _init(false),
  set(0)
{
  PROC;

  /* Check that the input expression is valid and set variables*/
    interpretExpression(_expr);
    
  /* Set up filtered array etc. */
    init();

  ENDPROC;
};

/* Destructor */
ImageRegion::~ImageRegion()
{
  PROC;

  delete _image;      	        _image                = 0;
  delete _fullimage;      	_fullimage            = 0;
  delete _poswcs;               _poswcs               = 0;
  delete _putils;               _putils               = 0;
  delete _regCentre;            _regCentre            = 0;
  // delete _array;            _array                    = 0;

 /* Close down opened files */
  //dataSetServer->close(_imageset);

  if (_init) {
     dataSetServer->close(_filteredset);
  }

  ENDPROC;

}

/* Method: init - for convenience put start-up code here */
void ImageRegion::init()
{
 PROC;

 /* Open the input image as a dataset */
   _imageset = dataSetServer->open( _imagefile, DataSet::Read );

 /* Map the array extension */
    if (_imagefile.find(":") != string::npos) {
       _array = _imageset->array(_imagefile);
    } else {
       _array = _imageset->array(_imagefile+":");
    }

 /* Read header keywords */
    readkeywords();

  /* Set the image pixel coordinates and units */
    setUnits();

 /* Get the region centre as a sourcePosition object so that it can
    be accessed in various coordinate systems */
   _putils = new PositionUtils(_imgAttribs->instrumentID());
   _poswcs = new PosWCS(_imageset, _imgAttribs->rapnt(), _imgAttribs->decpnt(),
                        _imgAttribs->papnt(), false);

  /* Check whether image axes are inverted */
    bool invert_axes = checkaxes();

  /* Make a pixel iterator from the input image */
    if (_imagefile.find(":") != string::npos) {
       _fullimage = new ImagePixelIterator(_imagefile, invert_axes);
    } else {
       _fullimage = new ImagePixelIterator(_imagefile+":", invert_axes);
    }

  /* Subset the input image using the expression to produce an array of
     pixels which are contained within the input region */

   _image = filterimage(_fullimage, _expr);

   errstr << "ImageRegion: constructed with "<<_expr
          << Sas::message(Msg::App, Msg::Verbose);

 /* Set internal extent and centre variables */
   setExtents();

   errstr<<" Setting up source pos: "<<_cenX<<" "<<_cenY<<" "<<_expCoordType
  	<< Sas::message(Msg::App, Msg::Verbose);

   _regCentre = new SourcePosition(_putils->
                           sourcePosition(_imageCoordType,
                                      _imgAttribs->datamode(),
                                      _imgAttribs->observationDate(),
                                      _cenX,
                                      _cenY,
                                      _poswcs));

 ENDPROC;
}

/* Method: srcCentre - returns the centre of the source region in the units
                       of the source expression */
CoordPair ImageRegion::centre() const
{
 PROC;

    RETURN CoordPair(_cenX, _cenY);
}

/* Method: centralCcd - returns the CCD containing the 
   centre of the source region in the units */
Ccd ImageRegion::centralCcd() const
{
 PROC;

    GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

    CoordPair pixel = conv2Det(centre());

    CamCoords testpixel = _putils->detxy005ToCamCoord2(pixel.x(),
                                                            pixel.y());
   /* Find which CCD this is on */
    Ccd ccd = ct->onCcd(testpixel.x(), testpixel.y());

    // delete ct; ct=0;

    RETURN ccd;
}

/* Method: conv2Det - converts any input coords into DETXY */
CoordPair ImageRegion::conv2Det(CoordPair input) const
{
 PROC;

    SourcePosition * sp = new SourcePosition(_putils->
                           sourcePosition(_imageCoordType,
                                      _imgAttribs->datamode(),
                                      _imgAttribs->observationDate(),
                                      input.x(),
                                      input.y(),
                                      _poswcs));

    double x = sp->xdet(), y = sp->ydet();

    delete sp; sp=0;

    RETURN CoordPair(x,y);
}

/* Method: conv2Pos - converts any input coords into XY */
CoordPair ImageRegion::conv2Pos(CoordPair input) const
{
 PROC;

    double xout=0, yout=0;

    if (_imageCoordType == "POS")
       RETURN input;

    else if (_imageCoordType == "DET") {

   /* Need to map the det and pos arrays from the filtered table */
       Table * table = _filteredset->table("EVENTS");
       Column * xColumn     = table->column("X");
       Column * yColumn     = table->column("Y");
       ColumnData * xColData     = xColumn->data();
       ColumnData * yColData     = yColumn->data();
       int32 * xData        = xColData->int32Data();
       int32 * yData        = yColData->int32Data();

   /* Loop over each region pixel */
       int i=0;
       FOREACH(_image)
       {
          PixelElement pixel = _image->currentItem();
       
          i++;

        // Does this pixel contain the input position ?
          if (abs(pixel.xCentre()-input.x()) < pixel.xWidth() &&
              abs(pixel.yCentre()-input.y()) < pixel.yWidth()) 
          {
             // Find the X,Y value for this pixel index
              xout=(double) xData[i];
              yout=(double) yData[i];

              //cout<<"IN: "<<xout<<" "<<yout<<" "<<pixel.xCentre()
                //  <<" "<<pixel.yCentre()<<" "<<pixel.xIndex()<<" "
                //<<" "<<pixel.yIndex()<<" "<<xData[pixel.xIndex()]<<endl;
              break;
          }
       }

       delete xColData, yColData;
    }

    RETURN CoordPair(xout,yout);
}

/* Method: conv2Raw - converts any input coords into RAW */
CoordPair ImageRegion::conv2Raw(CoordPair input) const
{
 PROC;

    CoordPair detpos = conv2Det(input);

    CamCoords camcoord2 = _putils->detxy005ToCamCoord2(
                                     INT(detpos.x()), INT(detpos.y()));

    PixCoords pixcoord = _putils->camCoord2ToPixCoord(camcoord2);

    RETURN CoordPair(pixcoord.x(), pixcoord.y());
}

/* Method: calcNormal - calculates the angle subtended by a tangent at
azimuthal angle, phi, to the optical axis (degrees) */
double ImageRegion::calcNormal(double phi) const
{
 PROC;

  /* Calculate the orientation (degs) of the ellipse wrt the optical axis */
   double orient = (phi * 180.0 / M_PI) - 90.0;

  // If not in detector coords need to convert using the position angle
   if (_expCoordType != "DET")
      orient = _imgAttribs->papnt() - (90.0 + orient);

   //double pa = _imgAttribs->papnt();

   //cout<<"Calc: "<<((phi * 180.0 / M_PI) - 90.0)<<" "<<pa<<" "<<orient;

   // DOESNT WORK FOR XY !!!!!

   RETURN orient;
}

/* Method: getCoordType - finds the coord type from the select expression */
const string ImageRegion::getCoordType(const string & expression)
{
 PROC;

   string ctype=" ";

   if (expression.find("DET") != string::npos) {
      ctype = "DET";
   } else if (expression.find("RA") != string::npos) {
      ctype = "EQPOS";
   } else {
      ctype = "POS";
   }

   RETURN ctype;
} 

/* Method: coordType - returns the coordinate type */
const string ImageRegion::coordType()
{
 PROC;

   RETURN _expCoordType;
}

/* Method: imcoordType - returns the coordinate type of the image */
const string ImageRegion::imcoordType()
{
 PROC;

   RETURN _imageCoordType;
}

/* Method: theta - returns the off-axis angle of this region */
double ImageRegion::theta() const
{
 PROC;

 RETURN _regCentre->theta();
}

/* Method: phi - returns the azimuthal angle of this region */
double ImageRegion::phi() const
{
 PROC;

 RETURN _regCentre->phi();
}

/* Method: celestialCentre - returns RA, DEC of source region centre*/
CoordPair ImageRegion::celestialCentre() const
{
 PROC;

   // Get centre in POS coords
    CoordPair posCentre = conv2Pos(centre());

   // Find RA,DEC at the centre
    CoordPair radec = _poswcs->wcs()->physicalCoords(posCentre.x(), 
                                                      posCentre.y());
    double ra = radec.x();
    double dec = radec.y();
    if (ra > 360.0)
	   ra = ra - 360.0;
    else if (ra < 0.0)
       ra = ra + 360.0;

    RETURN CoordPair(ra,dec);
 
 ENDPROC;
}

/* Centroid this region */
CoordPair ImageRegion::centroid(const int Niters ) 
{
  PROC;

  // This is a bit duff because it only uses the pixels within the
  // source region rather than the whole image

     double xpix=0.0, ypix=0.0;
     double total=0.0, offset=0.0;
     int ngood=0;
    
   /* Set initial centre value */
     double xcen = _cenX;
     double ycen = _cenY;

   /* Find the radius in image units */
     double arad = getRadius() * asecToImgUnits();

     for (int i=1; i<Niters+1; i++)
     {

      /* Initialise */
        total=0.0;
        ngood=0;
        xpix=0.0, ypix=0.0;

      /* Loop over each region pixel */
        FOREACH(_image)
        {
          PixelElement pixel = _image->currentItem();

       /* Include this pixel if it lies within the extraction radius */
          offset = sqrt( ( xcen - pixel.xCentre() )*( xcen - pixel.xCentre() ) 
                   + ( ycen - pixel.yCentre() )*( ycen - pixel.yCentre() ) );

          if (offset <= arad)
          {
            total = total+pixel.value();
            xpix = xpix + (pixel.xCentre() - xWidth()/2.0) * pixel.value();
            ypix = ypix + (pixel.yCentre() - yWidth()/2.0) * pixel.value();
            ngood = ngood + 1;
          }

        }

        if (ngood > 0) {
           xpix = xpix / total;
           ypix = ypix / total;
        }
        xcen = xpix;
        ycen = ypix;
     }

     errstr 
  	<< "Centroided on "<<xcen<<" "<<ycen
  	<< Sas::message(Msg::App, Msg::Verbose);

  RETURN CoordPair(xcen, ycen);
}

/* Method - counts() - returns the total counts in the image region */
double ImageRegion::counts() const
{
  PROC; 

     double total=0;

     FOREACH(_image)
     {
        PixelElement pixel = _image->currentItem();

        total = total+pixel.value();
     }

  RETURN total;
}

/* Method - ccdList() - returns a list of the CCDs included in the region */
vector<Ccd> ImageRegion::ccdList() const
{
  PROC; 

     GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);
     vector<Ccd> ccdlist(13);

     FOREACH(_image)
     {
        PixelElement inpixel = _image->currentItem();
        CoordPair pixel = conv2Det(CoordPair(inpixel.xCentre(),
                                             inpixel.yCentre()));
        CamCoords testpixel = _putils->detxy005ToCamCoord2(pixel.x(), 
                                                            pixel.y());

      /* Find which CCD this is on */
        Ccd ccd = ct->onCcd(testpixel.x(), testpixel.y());

      /* Set this CCD true in the list */
        ccdlist[ccd]=1;
     }

  RETURN ccdlist;
}

/* Method - centralPixelValue - returns the counts in the central pixel */
double ImageRegion::centralPixelValue() 
{
  PROC; 

     double value=0;
     //unsigned int i=0;

     FOREACH(_image)
     {
        //i++;

        PixelElement pixel = _image->currentItem();

        if (abs(pixel.xCentre() - _cenX) <= xWidth() &&
            abs(pixel.yCentre() - _cenY) <= yWidth() ) {
           value = pixel.value();
        }
        //if (i == _image->size() / 2 || i == (_image->size()+1)) {
          // PixelElement pixel = _image->currentItem();
           //value = pixel.value();
       // }
     }

  RETURN value;
}

/* Method flux() - returns the flux, mean counts per arcsec^2 */
double ImageRegion::flux()
{
  PROC; 

  RETURN (countsperpix() / pixelArea());
}

/* Method countsperpix() - returns the mean counts per pixel */
double ImageRegion::countsperpix()
{
  PROC; 

  RETURN (counts() / number_of_pixels());
}

/* Method number_of_pixels() - returns the number of pixels in the region */
double ImageRegion::number_of_pixels()
{
  PROC; 

  RETURN _image->size();
}

/* Method pixelArea() - returns the area of each image pixel in arcsecs */
double ImageRegion::pixelArea()
{
  PROC; 

  RETURN abs(xWidth() * yWidth()) * _xUnits * _yUnits;
}

/* Method area() - returns the area of the region in arcsec**2 */
double ImageRegion::area()
{
  PROC; 

  RETURN pixelArea() * number_of_pixels();
}

/* Method xWidth() - returns the xWidth of each image pixel in det/xy coords*/
double ImageRegion::xWidth()
{
  PROC; 

  RETURN _fullimage->wcsconvertor()->xWidth();
}

/* Method yWidth() - returns the yWidth of each image pixel in det/xy coords*/
double ImageRegion::yWidth()
{
  PROC; 

  RETURN _fullimage->wcsconvertor()->yWidth();
}

/* Method asecToImgUnits() - conversion factor between arcsec and image units */
double ImageRegion::asecToImgUnits()
{
  PROC; 

  RETURN (1.0 / _xUnits);
}

/* Method getRadius - returns the radius of the source region in arcsecs */
double ImageRegion::getRadius() const
{
  PROC; 

   /* Use the Y extents to calculate the radius - these dont suffer from
      the distortion of RA */

  RETURN (abs(_maxY - _minY) /2.0 * _yUnits);
}

/* Method exposure() - returns the exposure time for the image */
double ImageRegion::exposure() const
{
  PROC; 

  RETURN _exposure;
}

/* Method getRadProf - gets the radial profile from a position 
                       in the image out to a maximum radius */
ImageRegion::RadialProfile ImageRegion::getRadProf(double xcen, double ycen,
                                            double width, double maxrad) 
{
  PROC; 

      ImageRegion::RadialProfile r;

     /* Find the maximum radius to use in image units for a 5' radius */
      double maxoff = maxrad / _xUnits;
      double binwid = width / _xUnits;

     /* Set X axis values for radial profile */
      r.nbins = INT( maxoff / binwid ) + 1;  // Number of radial bins
      r.rwid = width;                    // Bin width (arcsec)

     /* Zero the data array */
      for (int i=0; i<r.nbins; i++) {
          r.data.push_back(0.0);
          r.error.push_back(0.0);
      }

//      cout<<"R: "<<r.nbins<<" "<<r.rwid<<" "<<binwid<<endl;

      vector<int> norm(r.nbins);

/* Loop over each element in the image (may need to subset first) */
      FOREACH(_fullimage)
      {
        PixelElement pixel = _fullimage->currentItem();

        double offset = sqrt( (pixel.xCentre()-xcen) * (pixel.xCentre()-xcen) 
                          + (pixel.yCentre()-ycen) * (pixel.yCentre()-ycen) );


      /* If pixel is within the required outer radius add to radprof bin */
      // Should also exclude if in chip gap etc...
        if (offset <= maxoff)
        {
           int bin = INT(offset / binwid);

           r.data[bin] += pixel.value();
           norm[bin]++;
        }
      }

    /* Normalise the output to per square arcsec */
      for (int i=0; i<r.nbins; i++) {
         if (norm[i] > 0) {
            r.error[i] = sqrt(r.data[i]) / (norm[i] * pixelArea());
            r.data[i] = r.data[i] / (norm[i] * pixelArea());
         }
//          cout<<"R3: "<<r.data[i]<<" "<<norm[i]<<" "<<pixelArea()<<endl;
      }

      
      RETURN r;
}


/* Method setExtents - find the X and Y limits of the spatial region */
const void ImageRegion::setExtents() 
{
  PROC; 

     _minX=100000000;
     _maxX=-100000000;
     _minY=100000000;
     _maxY=-100000000;

  /* Just find the max and min pixel positions in this array */
     FOREACH(_image)
     {
        PixelElement pixel = _image->currentItem();

        _minX = pixel.xCentre() < _minX ? pixel.xCentre():_minX;
        _maxX = pixel.xCentre() > _maxX ? pixel.xCentre():_maxX;
        _minY = pixel.yCentre() < _minY ? pixel.yCentre():_minY;
        _maxY = pixel.yCentre() > _maxY ? pixel.yCentre():_maxY;
     }

  /* Add in the extra half pixel */
     _minX = _minX - xWidth();
     _maxX = _maxX + xWidth();
     _minY = _minY - yWidth();
     _maxY = _maxY + yWidth();

     _cenX = (_minX + _maxX) / 2.0;
     _cenY = (_minY + _maxY) / 2.0;

  ENDPROC;
}
/* Method: filterimage - filters out pixels in the image which aren't
           contained within the selection expression */
TablePixelIterator * ImageRegion::filterimage(  ImagePixelIterator * fullimage, 
                                                   const string & expression )
{
  /* Create a container file to hold the temporary table */
    srand(getpid());
    bool filenotunique=true;
    string rawtablename    = "";
    while (filenotunique) {
       string postfix = toString(rand());
       rawtablename    = "rawpixellist" + postfix + ".ds:EVENTS";
       if (!dataSetServer->exists(rawtablename)) {
          set = dataSetServer->open(rawtablename, DataSet::Create);
          filenotunique=false;
       }
    }

  /* Subset the image array using the DSS extents */
  string Xaxis, Yaxis;
  double xmin, xmax, ymin, ymax;
  RangeBound lxb,uxb,lyb,uyb;

  if (_expCoordType == "DET") {
     Xaxis="DETX";
     Yaxis="DETY";
  }
  else if (_expCoordType == "EQPOS") {
     Xaxis="RA";
     Yaxis="DEC";
  } else {
     Xaxis="X";
     Yaxis="Y";
  }

 /* Set a DSS object */
  DataSubspace _dss = dssBuilder.getDss(expression);

  Range<double> extentX = _dss.realextent(Xaxis);
  Range<double> extentY = _dss.realextent(Yaxis);

  extentX.values(xmin, xmax, lxb,uxb);
  extentY.values(ymin, ymax, lyb,uyb);

  /* If the selection expression and image have the same coordinates we
     can define a subset of the image to use */

  string tempexp = expression; // Variable to hold expression which may change

  unsigned ixmin=0, ixmax=10000, iymin=0, iymax=10000;

  if ( (_expCoordType=="DET" && _imageCoordType=="DET") ||
       (_imageCoordType=="POS" && (_expCoordType=="POS" ||
                                  _expCoordType=="EQPOS")) ) 
  {

   /* If ranges selection is in POS coords, need to convert RA, DECS to X,Y */
   if (_expCoordType == "EQPOS") {

     double xpixwid_deg = xWidth()/ 72000.;
     double ypixwid_deg = yWidth()/ 72000.;

     double src_ra = (xmax+xmin) / 2.0;
     double src_dec = (ymax+ymin) / 2.0;
     int src_xwid_pix = (int)ceil(abs(src_ra - xmin) / xpixwid_deg);
     int src_ywid_pix = (int)ceil(abs(src_dec - ymin) / ypixwid_deg);

     errstr<<"Searching for closest pixel to source region centre"
        << Sas::message(Msg::App, Msg::Verbose);

    /* Find the image pixel which corresponds to the centre of the
       source region. Use this to set the range of image pixels to consider */
    fullimage->first();
    PixelElement pixel = fullimage->currentItem();
    CoordPair radecpt = _poswcs->wcs()->physicalCoords(pixel.xCentre(),
                                      pixel.yCentre());
    int ntries = 0;
    double cosdec = cos(src_dec * M_PI / 180.0);
    //double cosdec = cos(src_dec * 3.14159265 / 180.0);

   // Iterate until we find a pixel which contains the region centre
    double xdiff = (radecpt.x() - src_ra);
    if (xdiff < -180.0) xdiff +=360.0;
    if (xdiff > 180.0) xdiff -=360.0;

    while (abs(xdiff) > xpixwid_deg/cosdec ||
           abs(radecpt.y() - src_dec) > ypixwid_deg)
    {
     // Find the rough offset in pixels from the centre
      //int xdelta = (int)floor(cosdec * -(src_ra - radecpt.x()) / xpixwid_deg + 0.5);
      int xdelta = (int)floor(cosdec * xdiff / xpixwid_deg + 0.5);
      int ydelta = (int)floor(-(radecpt.y() - src_dec) / ypixwid_deg + 0.5);
      int ix = pixel.xIndex() + xdelta;
      int iy = pixel.yIndex() + ydelta;

      errstr<<"Tried: "<<ix<<" "<<iy<<" "<<radecpt.x()<<" "<<radecpt.y()
        <<" deltas: "<<xdelta<<" "<<ydelta
        << Sas::message(Msg::App, Msg::Noisy);

     // Find the position of this pixel
      fullimage->setIndex( (iy-1) * fullimage->xbins() + ix -1 );
      pixel = fullimage->currentItem();
      radecpt = _poswcs->wcs()->physicalCoords(pixel.xCentre(),
                                      pixel.yCentre());

      xdiff = (radecpt.x() - src_ra);
      if (xdiff < -180.0) xdiff +=360.0;
      if (xdiff > 180.0) xdiff -=360.0;

     // Check that we are getting there
      if (ntries++ > 20) { errstr<<"Region does not seem to be within "
                            <<"the image "<<Sas::error("InvalidRegion");}
    }

  /* Set the pixel range now we know the centre */
    ixmin=pixel.xIndex() - src_xwid_pix - 1;
    ixmax=pixel.xIndex() + src_xwid_pix + 1;
    iymin=pixel.yIndex() - src_ywid_pix - 1;
    iymax=pixel.yIndex() + src_ywid_pix + 1;

  /* If a simple circle or annulus convert the expression to X/Y */
    tempexp = convExp(expression,pixel.xCentre(),pixel.yCentre());

   } else {

    /* Convert the detector or sky coordinate ranges to array indices */
    ixmin = int( (xmin - fullimage->physXYmin().x()) / xWidth() - 1);
    ixmax = int( (xmax - fullimage->physXYmin().x()) / xWidth() + 1);
    iymin = int( (ymin - fullimage->physXYmin().y()) / yWidth() - 1);
    iymax = int( (ymax - fullimage->physXYmin().y()) / yWidth() + 1);
   }
  }

 /* Limit the range to be within the image */
  ixmin = ixmin < 1 ? 1 : ixmin;
  ixmax = ixmax > fullimage->xbins() ? fullimage->xbins() : ixmax;
  iymin = iymin < 1 ? 1 : iymin;
  iymax = iymax > fullimage->ybins() ? fullimage->ybins() : iymax;

  errstr<<"X : "<<xmin<<" "<<xmax<<" "<<ixmin<<" "<<ixmax
        <<"Y: "<<ymin<<" "<<ymax<<" "<<iymin<<" "<<iymax
        << Sas::message(Msg::App, Msg::Verbose);

  unsigned subsize = (ixmax - ixmin +1) * (iymax - iymin +1);

  /* Add the table */
  Table * table = set->addTable(rawtablename, subsize);

  /* Add columns to this table */
    Column * valueCol  = table->addColumn("VALUE",
                                        Column::Real64, "Value", "");
 
    Column * skyxCol    = table->addColumn("X",
                                         Column::Int32, "Sky X", "");

    Column * skyyCol    = table->addColumn("Y", Column::Int32, "Sky Y", "");

    Column * eqxCol    = table->addColumn("RA",
                                         Column::Real64, "RA", "");

    Column * eqyCol    = table->addColumn("DEC", Column::Real64, "DEC", "");

    Column * detxCol    = table->addColumn("DETX", Column::Int16, 
                                             "x det coords", "0.05 arcsec");

    Column * detyCol    = table->addColumn("DETY", Column::Int16, 
                                             "y det coords", "0.05 arcsec");

    Column * indexxCol    = table->addColumn("XPIXEL", Column::Int16,
                                         "x index", "");

    Column * indexyCol    = table->addColumn("YPIXEL", Column::Int16,
                                         "y index", "");

  Column * timeCol    = table->addColumn("TIME",
                                         Column::Real64,
                                         "elasped time",
                                         "sec");

  ColumnData * timeColData   = timeCol->data();
  real64     * timeData      = timeColData->real64Data();

    ColumnData * valueColData  = valueCol->data();
    real64     * valueData     = valueColData->real64Data();

  ColumnData * skyxColData   = skyxCol->data();
  int32      * skyxData      = skyxColData->int32Data();

  ColumnData * skyyColData   = skyyCol->data();
  int32      * skyyData      = skyyColData->int32Data();

  ColumnData * eqxColData   = eqxCol->data();
  real64     * eqxData      = eqxColData->real64Data();

  ColumnData * eqyColData   = eqyCol->data();
  real64     * eqyData      = eqyColData->real64Data();

  ColumnData * detxColData   = detxCol->data();
  int16      * detxData      = detxColData->int16Data();

  ColumnData * detyColData   = detyCol->data();
  int16      * detyData      = detyColData->int16Data();

  ColumnData * indexxColData   = indexxCol->data();
  int16      * indexxData      = indexxColData->int16Data();

  ColumnData * indexyColData   = indexyCol->data();
  int16      * indexyData      = indexyColData->int16Data();

  /* This conversion is HARDWIRED, but is probably acceptable
     as the DET coord definition is concrete. */
  double det2deg=1./(72000);

  /* Add WCS attributes to the DET columns (for attcalc) */
  WCSConvertor("DETX", 0, 0, det2deg, "DETY", 0, 0, det2deg).
                             addAttributesTo(detxCol, detyCol);

  /* Compute const time value to write to each row */
  double time = STime(_imageset->attribute("DATE-OBS")->asString())
                                            -Xmm::missionReferenceTime;

  /* Add WCS attributes to the index columns
     (for pixel-DET coord mapping; this way the WCS info
     can also store the pixel size) */
  fullimage->wcsconvertor()->addAttributesTo(indexxCol, indexyCol);

  /* Find the number of x bins */
  unsigned xbins = fullimage->xbins();

  /* Now iterate through input array adding a pixel entry for each */
  int row = 0;

  fullimage->first();

  /* Only access the elements of interest */
  for (unsigned iy = iymin; iy<=iymax; iy++)
  {
    fullimage->setIndex( (iy-1) * xbins + ixmin -1 );

    for (unsigned ix = ixmin; ix<=ixmax; ix++)
    {
      PixelElement pixel = fullimage->currentItem();
      timeData[row] = time;
      valueData[row]  = pixel.value();

     /* Set the X,Y and detector coordinates for each pixel */
      if (_imageCoordType == "POS") {

       /* Set the skypixel values */
         skyxData[row] = (int) pixel.xCentre();
         skyyData[row] = (int) pixel.yCentre();

       /* Derive the DETX,Y positions from the sky position */
         SourcePosition pixpos = _putils->sourcePosition("POS", 
                                 _imgAttribs->datamode(), 
                                 _imgAttribs->observationDate(),
                                 pixel.xCentre(), pixel.yCentre(),
                                      _poswcs);
         detxData[row] = (int) pixpos.xdet();
         detyData[row] = (int) pixpos.ydet();

       /* Find the celestial RA, DEC for these sky coordinates */
         CoordPair radecpt = _poswcs->wcs()->physicalCoords(pixel.xCentre(),
                                      pixel.yCentre());
         
         eqxData[row]   = radecpt.x();
         eqyData[row]   = radecpt.y();

      } else { // X,Y values will be derived later by attcalc */
         detxData[row] = (int) pixel.xCentre();
         detyData[row] = (int) pixel.yCentre();
      }

      indexxData[row] = pixel.xIndex();
      indexyData[row] = pixel.yIndex();
      ++row;

      errstr<<"added pixel: "<<pixel.xCentre()<<" "<<pixel.yCentre()
            <<"index["<<pixel.xIndex()<<","<<pixel.yIndex()<<"]"
                << Sas::message(Msg::App, Msg::Noisy);

     /* Move to next element if this isn't the last */
      if (! fullimage->isDone()) {
         fullimage->next();
      } 
    }
  }

  /* Close dataset  */
  delete indexyColData;
  delete indexxColData;
  delete detyColData;
  delete detxColData;
  delete skyyColData;
  delete skyxColData;
  delete valueColData;

  /* Run attcalc on this table if wanted */
  if (_imageCoordType == "DET") {
     runAttCalc(rawtablename);
  }

  dataSetServer->close(set);

  /* Clone this so that we can work with it */
   _filteredset = dataSetServer->clone( rawtablename,
                                      _filteredtablename,
                                      Dal::Temp);

  /* Remove the raw table */
   dataSetServer->remove(rawtablename);

  /* Create a data filter from the input expression */
   //DataFilter filter(expression);
   DataFilter filter(tempexp);

   table = _filteredset->table("EVENTS");
   
   long rows = filter.doFilter(*table);

   errstr << "Found " << rows << " in filtered table."
          << Sas::message(Msg::App, Msg::Verbose);

  /* Set a flag to say that the filtered table has been created */
   _init=true;

  /* Create a new table filter object pointing to this table */
        RETURN (new TablePixelIterator( _filteredtablename,
                                "XPIXEL",
                                "YPIXEL",
                                "VALUE"
                                ));
}

/* Method: convExp - convert simple RA,DEC expression into X/Y */
string ImageRegion::convExp( const string & expression, 
                               double xpix, double ypix )
{
  PROC;

   string outexp = expression;

   vector<double> circleRad;
   double CFAC = 3600.*20.; // Conversion from degrees to X/Y pix

  // Convert expression to a DSS object
   DataSubspace dss = dssBuilder.getDss(expression);

  // Check whether there are one or more region filters 
   if (dss.components() == 1 && dss.component(0)->filters()==1) 
    {
    // Convert the one region filter into a GeoRegion object 
     GeoRegion region = dss.component(0)->filter(0)->region();
     
    // Check that expression only contains one filter
     if (region[0][0]->type() == ShapeBasics::CIRCLE)
         {
             double radius=region[0][0]->radii()[0]*CFAC;
             outexp = "(X,Y) in CIRCLE(" + toString(xpix) + "," 
                      + toString(ypix) + "," + toString(radius) + ")";
         } 
     else if (region[0][0]->type() == ShapeBasics::ANNULUS)
         {
             double inner = region[0][0]->radii()[0]*CFAC;
             double outer = region[0][0]->radii()[1]*CFAC;
             outexp = "(X,Y) in ANNULUS(" + toString(xpix) + "," 
                      + toString(ypix) + ","  
                      + toString(inner) + "," + toString(outer) + ")";
         }
    }
   //cout<<"Expression converted to "<<outexp<<endl;

    RETURN outexp;
      
  ENDPROC;
}

/* Method: interpretExpression - read and interpret the selection expression */
void ImageRegion::interpretExpression( const string & expression )
{
  PROC;

  /* Check that the input expression is valid */
    if (! Evaluator::isValidExpression(expression)) {
       errstr<<"Invalid expression: "<<expression
             << Sas::error("InvalidExpression");
    }

 /* Find and set the coordinate system from the expression */
   _expCoordType = getCoordType(expression);

 /* Set a DSS object */
//   DataSubspace _dss = dssBuilder.getDss(expression);

  ENDPROC;
}
/* Method: setUnits - sets the units for the raw image pixels */
const void ImageRegion::setUnits()
{
  PROC;

   /* Attempt to get units from array extension */
   /* !!! Ignore this for the moment 
    if (_array->hasAttribute("CUNIT1"))
    {
       string sxunits = _array->attribute("CUNIT1")->asString();
       string syunits = _array->attribute("CUNIT2")->asString();

       if ( sxunits.find("deg") != string::npos) {
          _xUnits=3600.0;
          _yUnits=3600.0;
       }
    } else {
   */
     // Assume that these are detector or sky pixels
       _xUnits = 0.05;
       _yUnits = 0.05;
   // !!! }

    errstr<<" Units set to: "<<_xUnits<<" "<<_yUnits
          << Sas::message(Msg::App, Msg::Verbose);

   /* Set the image axes coordinates */
    if ( _array->attribute("CTYPE1")->asString().find("DET") != string::npos) 
    {
       _imageCoordType = "DET";
    } else {
       _imageCoordType = "POS";
    }

  ENDPROC;

}

const WCSConvertor * const ImageRegion::setSkyWCS()
{
  PROC;

      double cdelt1 = _array->attribute("CDELT1")->asReal() /
                             _imageset->attribute("REFXCDLT")->asReal();
      double cdelt2 = _array->attribute("CDELT2")->asReal() /
                             _imageset->attribute("REFYCDLT")->asReal();

      const WCSConvertor * _wcs = new WCSConvertor( "Pixel",
                           (unsigned) _array->attribute("CRPIX1")->asInt() ,
                           _imageset->attribute("REFXCRPX")->asReal() ,
                           cdelt1,"Pixel",
                           (unsigned) _array->attribute("CRPIX2")->asInt() ,
                           _imageset->attribute("REFYCRPX")->asReal() ,
                           cdelt2);
  RETURN _wcs;
}
void ImageRegion::runAttCalc(const string & rawtablename)
{
  PROC;

  /* Find the reference coordinates */
  CoordPair refcoords = _poswcs->wcs()->physicalRefCoords();

  /* Add attributes */
  * set->addAttribute("TELESCOP",  "mission name") = _imgAttribs->telescope();

  * set->addAttribute("INSTRUME",  "instrument name") = 
                                               _imgAttribs->instrument();

  * set->addAttribute("DATAMODE",  "instrument mode") = _imgAttribs->datamode();

  * set->addAttribute("DATE-OBS",  "Observation start date") = 
                                             _imgAttribs->observationDate();

  * set->addAttribute("DATE-END",  "Observation end date") = 
                                             _imgAttribs->observationDateEnd();

  * set->addAttribute("FILTER",    "Filter ID") = _imgAttribs->filter();

  * set->addAttribute("SUBMODE",      "Observing mode") =
                                           _imgAttribs->observingMode();


  * set->addAttribute("EMDHLOW",      "EMDHLOW") = _imgAttribs->emdhlow();
  * set->addAttribute("EMDHUPP",      "EMDHUPP") = _imgAttribs->emdhupp();

  /* The attcalc task needs the following to be set */
  * set->addAttribute("CCDID",        "") = 1;
  * set->addAttribute("CCDNODE",        "") = 0;
  * set->addAttribute("QUADRANT",        "") = 0;

  /* ATTCALC changes the CAL accuracy mode (and who knows what else)
     so must save and restore this afterwards */
  AccuracyLevel tempmode = calServer.state()->accuracyLevel()->get();

  /* Open the attcalc task, set the parameters and execute the task */
  Info::Task* attcalc = task("attcalc");
  ParameterInfo& attcalcpars = *attcalc->parameters();
  attcalcpars.revertToDefault();

  attcalcpars["eventset"]->setValue(rawtablename);
  attcalcpars["attitudelabel"]->setValue("fixed");
  attcalcpars["fixedra"]->setValue(_imgAttribs->rapnt());
  attcalcpars["fixeddec"]->setValue(_imgAttribs->decpnt());
  attcalcpars["fixedposangle"]->setValue(_imgAttribs->papnt());
  attcalcpars["refpointlabel"]->setValue("user");
  attcalcpars["nominalra"]->setValue(refcoords.x());
  attcalcpars["nominaldec"]->setValue(refcoords.y());
  attcalcpars["withatthkset"]->setValue(false);
  attcalc->exec();

  /* Restore the CAL accuracy mode */
  calServer.state()->accuracyLevel()->set(tempmode);

  ENDPROC;
}

void ImageRegion::readkeywords()
{
  PROC;

   /* Read in the standard keywords */
    _imgAttribs = new DatafileAttributes(_imagefile);

   /* Get the exposure */
    if (_array->hasAttribute("EXPOSURE"))
    {
       _exposure = _array->attribute("EXPOSURE")->asReal();
    }
    else
    {
       errstr << "Exposure keyword not found - setting to 1000 seconds "
              << Sas::message(Msg::App, Msg::Sparse);
       _exposure = 1000.0;
    }

  ENDPROC;
}

const bool ImageRegion::checkaxes()
{
  PROC;

   /* If the X axis name is "Y" or "DETY" or "DEC" then the image
      is inverted */
    if ( _array->attribute("CTYPE1")->asString().find("Y") != string::npos ||
         _array->attribute("CTYPE1")->asString().find("DEC") != string::npos ) 
    {
       ALTRETURNV( true );
    }
    RETURN false;
}

