// 
// Module: ImageRegion
// Author: Richard Saxton
//
// Description: 
//
// Given an input image, this class generates 
// a new PixelIterator for those pixels
// that are bounded by the specified region expression.
//
// Modifications: See ChangeLog
// 
// $Id: ImageRegion.h,v 1.11 2008/11/26 10:53:52 rsaxton Exp $

#ifndef ImageRegion_H
#define ImageRegion_H

// General classes
#include <iostream>

// SAS-specific classes
#include "PixelIterator.h"
#include "ImagePixelIterator.h"
#include "TablePixelIterator.h"
#include "PositionUtils.h"
#include "SourcePosition.h"
#include "DatafileAttributes.h"
#include "Xmm.h"

// General utility functions

// Local classes

class ImageRegion
{
public:

  /* Constructor */
  ImageRegion( const string & imagefile, const string & expression,
                        const string & tempfilename);

  /* Destructor */
  ~ImageRegion();

  /* Methods */
  void ImageRegion::init();
  void ImageRegion::interpretExpression(const string & expression);
  const void ImageRegion::setExtents();
  void ImageRegion::readkeywords();
  void ImageRegion::runAttCalc(const string & rawtablename);
  TablePixelIterator * ImageRegion::filterimage(  
                   ImagePixelIterator * fullimage, const string & expression );
  double ImageRegion::getRadius() const;
  double ImageRegion::exposure() const;
  double ImageRegion::counts() const;
  vector<Ccd> ImageRegion::ccdList() const;
  double ImageRegion::centralPixelValue();
  double ImageRegion::countsperpix();
  CoordPair ImageRegion::centroid(const int Niters );
  CoordPair ImageRegion::conv2Det(CoordPair c) const;
  CoordPair ImageRegion::conv2Raw(CoordPair c) const;
  CoordPair ImageRegion::conv2Pos(CoordPair c) const;
  CoordPair ImageRegion::celestialCentre() const;
  CoordPair ImageRegion::centre() const;
  Ccd ImageRegion::centralCcd() const;
  const string ImageRegion::getCoordType(const string & expression);
  const string ImageRegion::coordType();
  const string ImageRegion::imcoordType();
  double ImageRegion::theta() const;
  double ImageRegion::phi() const;
  double ImageRegion::flux();
  double ImageRegion::number_of_pixels();
  double ImageRegion::pixelArea();
  double ImageRegion::area();
  double ImageRegion::xWidth();
  double ImageRegion::yWidth();
  double ImageRegion::asecToImgUnits();
  double ImageRegion::calcNormal(double phi) const; 

  DatafileAttributes * ImageRegion::attributes() {return _imgAttribs;}

  /* Structure */
  struct RadialProfile{
        vector<double> data;     // The radial profile
        vector<double> error;     // The radial profile
        double   rmin,           // Minimum radius
                 rwid;           // Width of each radial bin (arcsec)
        int      nbins;          // Number of radial bins
  };

  /* Get a radial profile from an image 
     xcen, ycen are the profile centre in the units of the image
     width is the width of each radial profile bin in arcseconds */
  RadialProfile ImageRegion::getRadProf(double xcen, double ycen,
                                        double width, double maxrad);


  string ImageRegion::convExp(const string & posexp,
                                     double xCen, double yCen);

  const WCSConvertor * const ImageRegion::setSkyWCS();

protected:

  const void ImageRegion::setUnits();
  const bool ImageRegion::checkaxes();

private:

  DataSet * _filteredset;
  DatafileAttributes * _imgAttribs;
  string _expr;                        // The image region spatial expression
  string _filteredtablename; 
  string _imagefile;
  bool _init;
  DataSet * set;
  DataSet * _imageset;
  TablePixelIterator *  _image;
  double _minX;
  double _maxX;
  double _maxY;
  double _minY;
  double _cenX;
  double _cenY;
  ImagePixelIterator * _fullimage;
  const PosWCS *      _poswcs;
  PositionUtils * _putils;
  const SourcePosition * _regCentre;
  string _expCoordType;
  string _imageCoordType;
  Array * _array;
  double _xUnits;
  double _yUnits;
  string               _instrumentID;
  string               _telescope;
  string               _datamode;
  string               _dateobs;
  string               _dateend;
  Xmm::Instrument       _instr;
  int                  _emdhlow;
  int                  _emdhupp;

  double _rapnt;
  double _decpnt;
  double _papnt;
  double _exposure;

};

#endif
